/* =================================================================================== *
 * Color swatches:
 *
 * <div class=color-swatch>
 *     <div style="background: ...;"></div>
 *     <div style="background: ...;"></div>
 * </div>
 */

.color-swatch {
    display: flex;
    flex-wrap: wrap;
}

.color-swatch > div {
    height: 4em;
    width: 4em;
    border: black 0.7pt solid;
    display: flex;
    align-items: center;
    justify-content: center;
}

.color-swatch > hr {
    flex-basis: 100%;
    height: 0;
    margin: 0;
    border: none;
}

.small.color-swatch > div {
    height: 1em;
    width: 1em;
}

/* =================================================================================== *
 * Python-only and Rust-only labels: <i class=python-only>Python only!</i>
 */

.python-only, .tty-only, .rust-only, .gamut-only {
    font-size: 0.8em;
    display: inline-block;
    border-radius: 0.5em;
    padding: 0 0.6em;
    font-family: -apple-system, BlinkMacSystemFont, avenir next, avenir, segoe ui,
        helvetica neue, helvetica, Cantarell, Ubuntu, roboto, noto, arial, sans-serif;
    font-weight: 600;
    font-style: normal;
}

.python-only { background: #84c5fb; }
.tty-only    { background: #ffa9ff; }
.rust-only   { background: #f0ac84; }
.gamut-only  { background: #ddd; }
